
// coefficients contains the three main aerodynamic coefficients
// C_L, C_D and C_M, together with the angle of attack
// that they belong to.

// Written by Hans de Nivelle, November 2014.

#ifndef COEFFICIENTS_INCLUDED 
#define COEFFICIENTS_INCLUDED  1

#include <string>
#include <iostream>
#include <cmath>
#include "units.h"


struct coefficients
{
   // alpha always in radians!

   double alpha;   
   double lift;   
   double drag;  
   double moment;   // Around leading edge. Positive moment is always
                    // pitch up. This implies that moment is always <= 0.

   // axis = 0 denotes leading edge, axis = 1 denotes trailing edge.

   coefficients( double alpha, double lift, double drag, 
                 double axis, double moment )
      : alpha{ alpha }, lift{ lift }, drag{ drag },
        moment{ axis == 0.0 ? 
                   moment :
                   moment - axis * 
                        ( lift * cos( alpha ) + drag * sin( alpha )) }
   { }

   coefficients( double alpha, double lift, double drag,
                 double pressure_center ) 
      : coefficients( alpha, lift, drag, pressure_center, 0 ) 
   { }

   // Note that the following function may result division by zero,
   // when the force is aligned with the wing. This function
   // should not be used during flight. It is too unreliable: 

   double pressure_center( ) const
      { return - moment / ( lift * cos( alpha ) + drag * sin( alpha )); }

   // 0 is leading edge, 1 is trailing edge. The axis does not need
   // to be between 0 and 1. Don't forget the multiply the result
   // by the chord length of the wing. The stored value is
   // pitching moment around axis 0. 
  
   double relative_moment( double axis ) const
   {
      return moment + axis * ( lift * cos( alpha ) + drag * sin( alpha )); 
   }

   // Upside down means inverting everyting, accept the drag:

   coefficients upsidedown( ) const
   {
      return coefficients( -alpha, -lift, drag, 0.0, -moment );
   }

   // c1 and c2 must have identical alpha. Otherwise, the operation
   // fails.
   // Returns a balanced linear combination of c1 and c2. d = 0 means 
   // totally c1, while d = 1 means totally c2. 

   static coefficients interpolate( double d, const coefficients& c1,
                                              const coefficients& c2 )
   {
      if( c1. alpha != c2. alpha )
         throw std::runtime_error( "alphas differ" );

      return coefficients( c1. alpha,
                           ( 1 - d ) * c1. lift +   d * c2. lift,
                           ( 1 - d ) * c1. drag +   d * c2. drag,
                           0,
                           ( 1 - d ) * c1. moment + d * c2. moment ); 
   }
   
};

 
inline std::ostream&
operator << ( std::ostream& stream, const coefficients& c )
{
   stream << "alpha: " << c. alpha / 1.0_deg << " (deg)       ";
   stream << "lift: " << c. lift << "       ";
   stream << "drag: " << c. drag << "       ";
   stream << "moment: " << c. moment;
   return stream;
}


#endif


