
// Written by Hans de Nivelle, September 2014.

#ifndef LINALG_AFFINEMAP_INCLUDED
#define LINALG_AFFINEMAP_INCLUDED  1

#include "vector.h"
#include "matrix.h"


namespace linalg
{

   // An affine map is a function of form
   // F(x) = m.x + v, where m is a matrix, and v is a vector.  
   // We implemented this only for three dimensions. 
   // 
   // Affine maps can be used for scaling scenery objects.
   // 
   // Affine maps can be represented by 4X4 matrices as well, this
   // is the standard method in computer graphics, 
   // but using 4-vectors, collision detection is much harder.
   // Therefore, we do not use 4-vectors, except for direct interaction
   // with OpenGL.
   //
   // The position of an airplane, or any other vehicule, should be 
   // represented by a (proper) rigid transformation, or
   // rototranslation. 
   // A rototranslation is obtained when the matrix m is a rotation matrix.
   // We represent the matrix by a quaternion in this case, because
   // it is easier. 
 
   struct affinemap
   {
      matrix m; 
      vector v; 

      affinemap( ) 
         : m{ matrix::identity( ) }, 
           v{0,0,0} 
      { }

      affinemap( const matrix& m, const vector& v )
         : m{m},
           v{v}
      { }

      vector apply( const vector& x ) const { return m * x + v; } 
      vector operator( ) ( const vector& x ) const { return m * x + v; } 
         // Both operators do the same: Apply the map on x.

      affinemap operator * ( const affinemap& map ) const
         { return { m * map.m, m * map.v + v }; }
         // ( c1 * c2 ) * v = c1 * ( c2 * v ).
         // Note that c2 is carried out before c1.
         // One must be careful with this operator, because the outcome
         // is often counter intuitive.

      affinemap& composewith( const affinemap& map )
         { v = map.m(v) + map.v; m = map.m * m; return *this; } 
            //    (*this) := map * (*this). 
            // For every vector v, we have
            //    *newthis(v) = map( *oldthis( v )). 
   };

 
   inline 
   std::ostream& operator << ( std::ostream& stream, const affinemap& af )
   {
      stream << "Affine map: M = \n";
      stream << af. m;
      stream << "V = " << af. v << "\n";
      return stream;
   }
}

#endif


