
linalg::fullvector 
linalg::bandmatrix::apply_inv( fullvector v, double epsilon ) const
throw ( issingular ) 
{
   ASSERT( epsilon >= 0.0 );

   // We need a copy of ourselves that we will sweep:

   bandmatrix sweep = *this;

   // First we round small elements to zero: 

   for( unsigned int i = 0; i < getdimension( ); ++ i )
      sweep. round2zero( i, epsilon );

   // We count the columns that we could not sweep, because this tells
   // us the true dimension of the matrix:

   unsigned int failures = 0;

   // In the first stage, we construct the following invariant: 
   // Let D be the dimension of the matrix. When entering at level j,
   // we have, for each j' < j, exactly one row that has its first
   // nonzero element at position j'. (More precisely, the j' for which
   // we don't have such colum, are counted in failures.) 
   // We also have D-j rows that have only zeroes at positions j' < j.
 
   for( unsigned int j = 0; j < getdimension( ); ++ j )
   {

#if 0 
      std::cout << "--------------------------------\n";
      std::cout << "j = " << j << "\n";
      std::cout << "failures = " << failures << "\n"; 
      std::cout << sweep << "\n";
#endif

      // We first find the first row that has its first non-zero
      // element in column j. 

      unsigned int i0 = begin(j);
      unsigned int i1 = end(j);

      while( i0 < i1 && !first_nonzero_at( sweep, i0, j ))
         ++ i0; 

      // If we reached i1, then there was no such row.

      if( i0 == i1 )
      {
         failures = failures + 1;
      }
      else
      {
         // Now we look for the best i that we could sweep with.
         // It must have the highest absolute value on sweep [ best i ] [j], 
         // and it must not have nonzero elements that would fall out
         // of the band when cleaning row i0.

         unsigned int besti = i0;
         double maxabs = fabs( sweep [ i0 ] [j] );

         for( unsigned int i = i0 + 1; i < i1; ++ i )
         {
            if( first_nonzero_at( sweep, i, j ) && 
                zero_from( sweep, i, i0 + bandwidth + 1 )) 
            {
               double c = fabs( sweep[i][j] );
               if( c > maxabs )
               {
                  maxabs = c;
                  besti = i;
               }
            }
         }

#if 0 
         std::cout << "i0 = " << i0 << "\n"; 
         std::cout << "besti = " << besti << "\n";
         std::cout << "maxabs = " << maxabs << "\n";
         std::cout << "i1 = " << i1 << "\n";
#endif

         // Now that we have chosen, we can clear the other rows that 
         // start at j with our choice:

         for( unsigned int i = i0; i < i1; ++ i ) 
         {
            if( i != besti && first_nonzero_at( sweep, i, j ))
            {
               double c = sweep[i][j] / sweep[ besti ][j];
               sweep. subtract_rows( besti, c, i );
               sweep. round2zero( i, epsilon );  
                  // Round entries whose absolute value became less than
                  // epsilon. 

               v[i] -= c * v[ besti ];
            }
         }
#if 0
         std::cout << "after sweep " << sweep << "\n";
         std::cout << "v = " << v << "\n";
#endif
      }
   }

   if( failures > 0 )
      throw issingular( dimension - failures, dimension );

   // Now we have for every j, one row that starts in column j.

   linalg::fullvector result = linalg::fullvector( dimension );

   unsigned int j = dimension;
   while(j)
   {
      -- j;

      unsigned int i0 = sweep. begin(j);
      unsigned int i1 = sweep. end(j);

      // We find the unique row istart that starts at position j.
      // It is guaranteed to exist, because of the invariant of the previous
      // loop.  

      unsigned int istart = i0; 
      while( !first_nonzero_at( sweep, istart, j ))
         ++ istart;

      ASSERT( istart < i1 );
         
      // This is an invariant check. If you don't want it, you can remove 
      // it. There should be no other rows that start at position j: 

      for( unsigned int i = istart + i; i < i1; ++ i )
      {
         ASSERT( !first_nonzero_at( sweep, i, j ));
      }

      // Using row istart, we clean the other rows:

      for( unsigned int i = i0; i < i1; ++ i ) 
      {
         if( i != istart )
         { 
            double c = sweep[i][j] / sweep[ istart ][j];
            sweep. subtract_rows( istart, c, i );
            sweep. round2zero( i, epsilon ); 
            v[i] -= c * v[ istart ];
         }
      }

      v[ istart ] /= sweep[ istart ][j]; 
      sweep [ istart ][j] = 1.0;  

      result [j] = v [ istart ];
   }

   return result; 
}


