
#ifndef LINALG_BANDMATRIX_INCLUDED
#define LINALG_BANDMATRIX_INCLUDED  1

// Written by Hans de Nivelle, 04.04.2012
// Made polymorphic on 17.10.2014.

#include <vector>
#include <iostream>
#include "issingular.h"
#include "generalvector.h"

namespace linalg
{
   // In order to use type X in a bandmatrix, you must
   // have an appropriate definition of norm<X>. 

   // A diagonal matrix has bandwidth 0.
   // If you have 3 on each row, then the bandwidth is 1.
   //             5                                    2.

   template< typename X, size_t B > 
   class bandmatrix
   {
      size_t dim;

      std::vector<X> repr;
         // Size of the vector is dim * ( 2*B + 1 ). 
         // We store the bandmatrix row wise. A matrix with a bandwidth
         // of 1, will have 3 elements in each row. They are stored
         // as follows: 
         // ( X{}, m[0][0], m[0][1], m[1][0], m[1][1], m[1][2], 
         //                          m[2][1], m[2][2], m[2][3], ...

 
      struct row_ref 
      {
         typename std::vector<X> :: iterator p; 
         size_t j0;
         size_t j1;         

         row_ref( typename std::vector<X> :: iterator p,
                  size_t j0, size_t j1 )
            : p{p}, 
              j0{j0}, j1{j1}
         { }

         X& operator [] ( size_t j ) const
         {
            if( j < j0 || j >= j1 )
               throw std::runtime_error( "row index out of band" );
            return p[ j - j0 ];
         }
      };


      struct const_row_ref 
      {
         typename std::vector<X> :: const_iterator p;  
         size_t j0;
         size_t j1;

         const_row_ref( typename std::vector<X> :: const_iterator p,
                        size_t j0, size_t j1 )
            : p{p}, 
              j0{j0}, j1{j1}
         { }

         X operator [] ( size_t j ) const
         { 
            if( j < j0 || j >= j1 )
               throw std::runtime_error( "row index out of band" ); 
            return p[ j - j0 ];
         }
      };             

   public: 
      bandmatrix( ) = delete;

      explicit bandmatrix( size_t dimension, 
                           const X& x = normdef<X> { }. zero( ))
         : dim{ dimension },
           repr{ std::vector<X> ( dimension * ( 2*B + 1 ), x )}
      { 
         if( dimension <= B ) 
            throw std::runtime_error(
               "dimension must be at least equal to bandwidth" );
      }


      // one should always write expressions of form 
      // m[i][j], never write m[i] in isolation. 

      row_ref operator [] ( size_t i );
      const_row_ref operator [] ( size_t i ) const; 

      // Given one of the indices, begin(i) and end(i) specify in
      // which range the other index must be. So one would have 
      // begin(i) <= j < end(i).

      size_t begin( size_t i ) const 
         { if( i >= dim ) throw std::runtime_error("column is out of range");
           return ( i >= B ) ? ( i - B ) : 0; }
      size_t end( size_t i ) const 
         { if( i >= dim ) throw std::runtime_error("column is out of range"); 
           return ( i + B < dim ) ? ( i + B + 1 ) : dim; }

      size_t dimension( ) const { return dim; }
      constexpr size_t bandwidth( ) { return B; }

      generalvector<X> operator( ) ( const generalvector<X> & v ) const;

      generalvector<X> apply_inverse( generalvector<X> v, 
                                      double epsilon ) const;
         // Apply our inverse on v. Constructing the inverse is not
         // possible, because it will not be a bandmatrix.
         // apply_inverse( ) may throw an issingular exception.
         // I tested it on double, math::complex and linalg::quaternion.
 
      bandmatrix<X,B> operator - ( ) const
      {
         auto res = *this;
         for( X& x : res. repr ) x = -x;
         return res;
      }

      template< typename D > 
      bandmatrix<X,B> & operator *= ( D d )
         { for( X& x : repr ) x *= d; return *this; } 

      template< typename D > 
      bandmatrix<X,B> & operator /= ( D d )
         { for( X& x : repr ) x /= d; return *this; }

      bandmatrix<X,B> & operator += ( const bandmatrix<X,B> & b )
         {  if( repr. size( ) != b. repr. size( ))
               throw std::runtime_error( "+= : dimensions differ" );
            for( size_t i = 0; i < repr. size( ); ++ i )
               repr [i] += b. repr [i];
            return *this; 
         }

      bandmatrix<X,B> & operator -= ( const bandmatrix<X,B> & b )
         {  if( repr. size( ) != b. repr. size( ))
               throw std::runtime_error( "-= : dimensions differ" );
            for( size_t i = 0; i < repr. size( ); ++ i )
               repr [i] -= b. repr [i];
            return *this; 
         }

#if 0
   private: 
      // These are all helper functions, related to applying
      // the inverse. 

#endif 

      void round2zero( size_t i, double epsilon );
         // In row i, make the elements whose norm is less 
         // than epsilon, equal to zero. 

      void subtract_rows( size_t i_1, X x, size_t i2 );
         // Subtract x * i1-th row from i2-th row. 

      static size_t greater( size_t i1, size_t i2 )
         { return i1 < i2 ? i2 : i1; }
      static size_t smaller( size_t i1, size_t i2 )
         { return i1 < i2 ? i1 : i2; }

      bool first_nonzero_at( size_t i, size_t j ) const;
         // True if first non-zero element in row j is at
         // column i.

      bool zero_from( size_t i, size_t j ) const;
         // True if row j contains zero at column i,
         // and in all columns i' > i. 

      void print ( std::ostream& ) const; 
   };


   template< typename X, size_t B >
   void bandmatrix<X,B> :: print( std::ostream& out ) const
   {
      out << "bandmatrix<" << B << "> [" << dim << "]:\n";

      for( size_t i = 0; i < dim; ++ i )
      {
         out << i << ":   ";
         for( auto j = begin(i); j < end(i); ++ j )
         {
            out << j << ":" << (*this)[i][j] << "   ";
         }
         out << "\n";
      }
      out << "\n";
   }


   template < typename X, size_t B >
   inline std::ostream& operator << ( std::ostream& out, 
                                      const bandmatrix<X,B> & b )
   {
      b. print( out );
      return out; 
   }


   template< typename X, size_t B > 
   typename bandmatrix<X,B>::const_row_ref
   bandmatrix<X,B>::operator [] ( size_t i ) const
   {
      if( i >= dim ) throw std::runtime_error( "i >= dimension" ); 

      // [ j0 .. j1 ) is the range in which we will accept the second index.
      // It is normally [ j - bandwidth .. j + bandwidth + 1 ), unless
      // one of them is out of range. The first should be not less than
      // 0, and the second should be not greater than the dimension of the
      // matrix.

      auto j0 = begin(i);
      auto j1 = end(i);

      typename std::vector<X> :: const_iterator p =
          repr. begin( ) + i * ( 2*B + 1 );

      if( i < B ) p += ( B - i ); 

      return const_row_ref( p, j0, j1 );
   }


   template< typename X, size_t B > 
   typename bandmatrix<X,B>::row_ref
   bandmatrix<X,B>::operator [] ( size_t i )
   {
      if( i >= dim ) throw std::runtime_error( "i >= dimension" ); 

      // [ j0 .. j1 ) is the range in which we will accept the second index.
      // It is normally [ j - bandwidth .. j + bandwidth + 1 ), unless
      // one of them is out of range. The first should be not less than
      // 0, and the second should be not greater than the dimension of the
      // matrix.

      auto j0 = begin(i); 
      auto j1 = end(i); 
   
      typename std::vector<X> :: iterator p = repr. begin( ) + i * (2*B + 1);

      if( i < B ) p += ( B - i );

      return row_ref( p, j0, j1 );
   }

   template< typename X, size_t B >
   generalvector<X> 
   operator * ( const bandmatrix<X,B> & m, const generalvector<X> & v ) 
   {
      if( m. dimension( ) != v. dimension( ))
         throw std::runtime_error( "bandmatrix * vector: dimensions differ" );

      generalvector<X> res = generalvector<X> ( v. dimension( ));

      for( size_t i = 0; i < v. dimension( ); ++ i )
         for( size_t j = m. begin(i); j != m. end(i); ++ j )
            res[i] += m[i][j] * v[j];
      return res; 
   }

   template< typename X, size_t B >
   generalvector<X>
   operator * ( const generalvector<X> & v, const bandmatrix<X,B> & m ) 
   {
      if( v. dimension( ) != m. dimension( ))
         throw std::runtime_error( "vector * bandmatrix: dimensions differ" );

      generalvector<X> res = generalvector<X> ( v. dimension( ));

      for( size_t j = 0; j < v. dimension( ); ++ j )
         for( size_t i = m. begin(j); i != m.end(j); ++ i )
            res[j] += v[i] * m[i][j];
      return res;
   }


   template< typename X, size_t B >
   inline generalvector<X> 
   bandmatrix<X,B> :: operator( ) ( const generalvector<X> & v ) const
      { return (*this) * v; } 


   template< typename X, size_t B, typename D >
   inline bandmatrix<X,B> operator * ( bandmatrix<X,B> m, D d )  
   {
      m *= d;
      return m;
   }

   template< typename X, size_t B, typename D >
   inline bandmatrix<X,B> operator * ( D d, bandmatrix<X,B> m ) 
   {
      m *= d;
      return m;
   }

   template< typename X, size_t B, typename D >
   inline bandmatrix<X,B> operator / ( bandmatrix<X,B> m, D d )
   {
      m /= d;
      return m;
   }

   template< typename X, size_t B >
   inline bandmatrix<X,B> operator + ( bandmatrix<X,B> m1, 
                                       const bandmatrix<X,B> & m2 )
   {
      m1 += m2;
      return m1;
   }

   template< typename X, size_t B >
   inline bandmatrix<X,B> operator - ( bandmatrix<X,B> m1,
                                       const bandmatrix<X,B> & m2 )
   {
      m1 -= m2;
      return m1;
   }


   /////////////////////////////////////////////////////////
   // Things related to apply_inverse(   ).
   /////////////////////////////////////////////////////////

   template< typename X, size_t B >
   inline void bandmatrix<X,B> :: round2zero( size_t i, double epsilon )
   {
      normdef<X> def;   // Definition of norm and zero. 
      
      auto j0 = begin(i);
      auto j1 = end(i);

      for( auto j = j0; j < j1; ++ j )
      {
         if( def( (*this) [i] [j] ) < epsilon )
            (*this) [i] [j] = def. zero( ); 
      }
   }

   template< typename X, size_t B >
   inline void bandmatrix<X,B> :: subtract_rows( size_t i1, X x, size_t i2 )
   {
      auto j0 = greater( begin(i1), begin(i2) );
      auto j1 = smaller( end(i1), end(i2) );

      for( auto j = j0; j < j1; ++ j )
         (*this) [i2][j] -= x * (*this) [i1][j];
   } 

   template< typename X, size_t B >
   inline bool bandmatrix<X,B> :: first_nonzero_at( size_t i, size_t j ) const
   {
      normdef<X> def; 

      if( def( (*this)[i][j] ) == 0 ) return false;

      auto j0 = begin(i);
      while( j > j0 )
      {
         -- j;
         if( def( (*this)[i][j] ) != 0 ) return false;
      }
      return true;
   }

   template< typename X, size_t B >
   inline bool bandmatrix<X,B> :: zero_from( size_t i, size_t j ) const
   {
      normdef<X> def;

      auto j1 = end(i);
      while( j < j1 )
      {
         if( def( (*this)[i][j] ) != 0 ) return false;
         ++ j;
      }
      return true;
   }

   template< typename X, size_t B > 
   generalvector<X> 
   bandmatrix<X,B> :: apply_inverse( generalvector<X> v,
                                     double epsilon ) const
   {
      if( epsilon < 0.0 )
         throw std::runtime_error( "apply_inverse: epsilon must be >= 0" );

      if( dimension( ) != v. dimension( ))
         throw std::runtime_error( "apply_inverse: dimensions differ" );

      normdef<X> def;  // Needed from computing norms.

      // We need a copy of ourselves that we will sweep:

      bandmatrix<X,B> sweep = *this;

      // First we round small elements to zero: 

      for( size_t i = 0; i < dimension( ); ++ i )
         sweep. round2zero( i, epsilon );

      // We count the columns that we could not sweep, because this tells
      // us the true dimension of the matrix:

      size_t failures = 0;

      // In the first stage, we construct the following invariant: 
      // Let D be the dimension of the matrix. When entering at level j,
      // we have, for each j' < j, exactly one row that has its first
      // nonzero element at position j'. (More precisely, the j' for which
      // we don't have such colum, are counted in failures.) 
      // We also have D-j rows that have only zeroes at positions j' < j.
 
      for( size_t j = 0; j < dimension( ); ++ j )
      {
#if 0 
         std::cout << "--------------------------------\n";
         std::cout << "j = " << j << "\n";
         std::cout << "failures = " << failures << "\n"; 
         std::cout << sweep << "\n";
#endif

         // We first find the first row that has its first non-zero
         // element in column j. 

         auto i0 = begin(j);
         auto i1 = end(j);

         while( i0 < i1 && ! sweep. first_nonzero_at( i0, j ))
            ++ i0; 

         // If we reached i1, then there was no such row.
   
         if( i0 == i1 )
            failures = failures + 1;
         else
         {
            // Now we look for the best i that we could sweep with.
            // It must have the highest absolute value on sweep [ best i ] [j], 
            // and it must not have nonzero elements that would fall out
            // of the band when cleaning row i0.

            auto besti = i0;
            double maxnorm = def( sweep [ i0 ] [j] );

            for( auto i = i0 + 1; i < i1; ++ i )
            {
               if( sweep. first_nonzero_at( i, j ) && 
                   sweep. zero_from( i, i0 + B + 1 )) 
               {
                  double n = def( sweep[i][j] );
                  if( n > maxnorm )
                  {
                     maxnorm = n;
                     besti = i;
                  }
               }
            }

#if 0 
            std::cout << "i0 = " << i0 << "\n"; 
            std::cout << "besti = " << besti << "\n";
            std::cout << "maxnorm = " << maxnorm << "\n";
            std::cout << "i1 = " << i1 << "\n";
#endif

            // Now that we have chosen, we can clear the other rows that 
            // start at j with our choice:

            for( auto i = i0; i < i1; ++ i ) 
            {
               if( i != besti && sweep. first_nonzero_at( i, j ))
               {
                  X x = sweep[i][j] / sweep[ besti ][j];
                  sweep. subtract_rows( besti, x, i );
                  sweep. round2zero( i, epsilon );  
                     // Round entries whose norm became less than
                     // epsilon. 

                  v[i] -= x * v[ besti ];
               }
            }
#if 0
            std::cout << "after sweep: " << sweep << "\n";
            std::cout << "v = " << v << "\n";
#endif 
         }
      }

      if( failures > 0 )
         throw issingular( dimension( ) - failures, dimension( ));

      // Now we have for every j, one row that starts in column j.

      generalvector<X> result = generalvector<X> ( dimension( ) );

      auto j = dimension( );
      while(j)
      {
         -- j;

         auto i0 = sweep. begin(j);
         auto i1 = sweep. end(j);

         // We find the unique row istart that starts at position j.
         // It is guaranteed to exist, because of the invariant of the previous
         // loop.  

         auto istart = i0; 
         while( ! sweep. first_nonzero_at( istart, j ))
            ++ istart;

         if( istart >= i1 )
            throw std::out_of_range( "one should have: istart < i1" );  
         
         // This is an invariant check. If you don't like it, you can remove 
         // it. There should be no other rows that have first non-zero
         // entry at position j: 

#if 0
         std::cout << "j = " << j << "\n";
         std::cout << "istart = " << istart << "\n";
         std::cout << sweep << "\n";
#endif

         for( auto i = istart + 1; i < i1; ++ i )
         {
            if( sweep. first_nonzero_at( i, j ))
               throw std::out_of_range( 
                  "There are two rows starting at position j" );
         }

         // Using row istart, we now clean the other rows:

         for( auto i = i0; i < i1; ++ i ) 
         {
            if( i != istart )
            { 
               X x = sweep[i][j] / sweep[ istart ][j];
               sweep. subtract_rows( istart, x, i );
               sweep. round2zero( i, epsilon ); 
               v[i] -= x * v[ istart ];
            }
         }

         v[ istart ] = ( 1.0 / sweep[ istart ][j] ) * v [ istart ];
            // This matters for quaternions.

         result [j] = v [ istart ];
      }
      return result; 
   }
} 

#endif

