
#ifndef LINALG_FOURMATRIX_INCLUDED
#define LINALG_FOURMATRIX_INCLUDED   1

// Written by Hans de Nivelle, December 2010. 
// A fourmatrix is a homogeneous matrix.
// We do not use homogeneous matrices in our flight-simulation project, 
// but they are needed in the interface to OpenGL. 

#include <iostream>

#include "fourvector.h"
#include "affinemap.h"

namespace linalg
{

   class fourmatrix 
   {
      double repr [4][4];

   public: 
      fourmatrix( );
         // Builds identity matrix.

      fourmatrix( const matrix& m );
         // Builds 4X4 matrix with same effect as m. 

      explicit fourmatrix( const vector& v );
         // Builds a translation over vector v.

      fourmatrix( const affinemap& map );
         // Builds 4X4 matrix with same effect as map.

      static fourmatrix frustum( double x0, double x1,
                                 double y0, double y1, 
                                 double z0, double z1 );
         // Builds a matrix that represents a frustum projection.
         // Note that z0 and z1 must be negative, and that one should have
         // z0 > z1, because we look down along the negative axis. 
         // 
         // x0 .. x1 and y0 .. y1 determine the size of the viewing window,
         // as it will appear at a distance of 1 in front of us. This is 
         // different than the OpenGL interface, but I think it is more
         // easy to use. 
         // 
         // The point [ x0, y0, z0 ] will be mapped to [ 1/z0, 1/z0, -1 ].
         // (Note that z0 is negative, so that it will appear bottom left.)
         // The point [ x1, y1, z1 ] will be mapped to [ -1/z1, -1/z1, 1 ].
         // (Because z1 is negative, this will appear top right.) 
         // Also note that the meaning of Z is inversed. After the 
         // transformation, a smaller Z means closer to us. 

      static fourmatrix orthographic( double x0, double x1, 
                                      double y0, double y1,
                                      double z0, double z1 );
         // Builds a matrix that represents an orthographic projection.
         // Note that z0 and z1 must be negative, and that one should have
         // z0 > z1. 
         // The matrix maps [ x0, y0, z0 ] to [ -1, -1, -1 ].
         // It maps [ x1, y1, z1 ] to [ 1, 1, 1 ]. 

      fourvector apply( const fourvector& v ) const;          
      fourvector operator( ) ( const fourvector& v ) const 
         { return apply(v); } 

      fourmatrix& composewith( const fourmatrix& second );
         // Left compose *this with second.
         // *this := second * (*this). For every vector v we will have
         // *newthis(v) = second( *oldthis(v) ).
 
      friend fourmatrix operator * ( const fourmatrix& , const fourmatrix& ); 
      friend std::ostream& operator << ( std::ostream & , const fourmatrix& );
      friend double* & operator << ( double* & , const fourmatrix& ); 

   };

   fourmatrix operator * ( const fourmatrix& m1, const fourmatrix& m2 );

   std::ostream& operator << ( std::ostream& stream, const fourmatrix& m ); 

   double* & operator << ( double* & p, const fourmatrix& m );
      // Prints the matrix into the pointer p, in the order
      // m[0][0], m[1][0], m[2][0], m[3][0], m[1][0], ...
      // The result can be loaded into OpenGL with
      // glLoadMatrixd( ).

   inline fourmatrix& fourmatrix::composewith( const fourmatrix& second )
      { *this = second * (*this); return *this; } 

}

#endif

