
// A generalvector is a vector of arbitrary length, containing
// elements of arbitrary type. Normally the type would be
// double or complex. Quaternion should be also possible but 
// it may cause problems due to the fact quaternion multiplication is 
// non-commutative. 

#ifndef LINALG_GENERALVECTOR_INCLUDED
#define LINALG_GENERALVECTOR_INCLUDED  1

#include <iostream>
#include <vector>
#include <string>
#include "normdef.h"

namespace linalg
{

   template< typename X >
   class generalvector 
   {
      std::vector<X> elements;

   public: 
      // Create a generalvector of size dimension, filled with
      // the given values, or default zeroes, as defined by normdef<X>.

      explicit generalvector( size_t dimension, 
                              const X& x = normdef<X> { }. zero( ))
         : elements{ std::vector<X> ( dimension, x ) }
      { }


      generalvector( std::initializer_list<X> values )
         : elements{ std::vector<X> ( values ) }
      { }
      
      generalvector( const std::vector<X> & values )
         : elements{ values }
      { }

      generalvector( std::vector<X> && values )
         : elements{ std::move( values ) }
      { }

  
      X& operator [] ( size_t i )
         { return elements[i]; }

      X operator [] ( size_t i ) const
         { return elements[i]; }

      size_t dimension( ) const { return elements. size( ); } 

      double norm( ) const 
      {
         normdef<X> def;  // Definition of norm for X.

         double n = 0;
         for( X x : elements ) 
            n += def(x);
         return n;
      }


      generalvector& operator += ( const generalvector<X> & v )
      {
         if( elements. size( ) != v. elements. size( ))
            throw std::runtime_error( "+= : dimensions of arguments differ" );
         for( size_t i = 0; i < elements. size( ); ++ i )
            elements[i] += v. elements [i];
         return *this;  
      }

      generalvector& operator -= ( const generalvector<X> & v )
      {
         if( elements. size( ) != v. elements. size( ))
            throw std::runtime_error( "-= : dimensions of arguments differ" );
         for( size_t i = 0; i < elements. size( ); ++ i )
            elements[i] -= v. elements [i];
         return *this; 
      } 


      // D can be everything on the condition that
      // X *= D is defined. One must be a bit careful when * is not
      // commutative. It is assumed that D is to the right. 

      template< typename D >
      generalvector& operator *= ( D d )
      {
         for( auto p = elements. begin( ); p != elements. end( ); ++ p )
            (*p) *= d;
         return *this;
      }

      // X /= D must be defined. 

      template< typename D >
      generalvector& operator /= ( D d )
      {
         for( auto p = elements. begin( ); p != elements. end( ); ++ p )
            (*p) /= d;
         return *this;
      }

      // Since X*D need not be commutative, we give two multiplication 
      // operators that are explicit about the side.

      template< typename D >
      generalvector& leftmultiply( D d )
      {
         for( auto p = elements. begin( ); p != elements. end( ); ++ p )
            (*p) = d * (*p);
         return *this;
      }

      template< typename D >
      generalvector& rightmultiply( D d )
      {
         for( auto p = elements. begin( ); p != elements. end( ); ++ p )
            (*p) = (*p) * d;
         return *this;
      } 

      generalvector operator - ( ) const
      {
         generalvector<X> res = *this;
         for( auto p = res. elements. begin( ); 
                   p != res. elements. end( ); ++ p )
            (*p) = - (*p);
         return res;
      }
 
      void print( std::ostream& out ) const
      {
         out << "{";
         for( auto p = elements. cbegin( ); p != elements. cend( ); ++ p )
         {
            if( p == elements. cbegin( ))
               out << " ";
            else
               out << ", "; 
            out << *p;
         }
         out << " }"; 
      }
   };


   template< typename X >
   inline generalvector<X> 
   operator + ( generalvector<X> v1, const generalvector<X> & v2 )
   {
      v1 += v2;
      return v1;
   }

   template< typename X >
   inline generalvector<X>
   operator - ( generalvector<X> v1, const generalvector<X> & v2 )
   {
      v1 -= v2;
      return v1;
   }

   template< typename X, typename D >
   inline generalvector<X> 
   operator * ( generalvector<X> v, D d )
   {
      v. rightmultiply(d);
      return v;
   }

   template< typename X, typename D >
   inline generalvector<X> 
   operator * ( D d, generalvector<X> v )
   {
      v. leftmultiply(d);
      return v;
   }

   template< typename X, typename D >
   inline generalvector<X>
   operator / ( generalvector<X> v, D d )
   {
      v /= d;
      return v;
   }

   // Multiplication of a vector with a vector is dot product:

   template< typename X >
   X operator * ( const generalvector<X> & v1, const generalvector<X> & v2 )
   {
      if( v1. dimension( ) != v2. dimension( ))
         throw std::runtime_error( "vector * vector: dimensions differ" );

      X res = normdef<X> ( ). zero( ); 

      for( size_t i = 0; i < v1. dimension( ); ++ i )
         res += v1[i] * v2[i];
      return res; 
   }

   template< typename X >
   inline std::ostream& operator << ( std::ostream& out, 
                                      const generalvector<X> & v )
   {
      v. print( out );
      return out;
   }
} 

#endif

