
// Written by Hans de Nivelle, Jan 2017.

#ifndef PHYSICS_RIGID_INCLUDED
#define PHYSICS_RIGID_INCLUDED  1

#include <vector> 
#include <initializer_list> 
#include <iostream>

#include "force_torque.h" 
#include "state.h" 
#include "../linalg/matrix.h"


namespace physics
{

   struct pointmass 
   {
      double mass;
      linalg::vector position;

      pointmass( ) = delete; 
      pointmass( double mass, linalg::vector position )
         : mass{ mass },
           position{ position }
      { }

      linalg::matrix inertia_matrix( const linalg::vector ref ) const;
         // Inertial matrix relative to given reference point.
   };

   inline std::ostream& operator << ( std::ostream& out, pointmass p )
   {
      return out << p.mass << " @ " << p. position;
   }


   struct mass_distribution
   {
      double total;
      linalg::vector center;    
      linalg::matrix inertia;   // Relative to center in internal
                                // coordinates.

      mass_distribution( double total, const linalg::vector& center,
                         const linalg::matrix& inertia )
         : total{ total },
           center{ center },
           inertia{ inertia }
      { }

      mass_distribution( const std::vector< pointmass > & distr );
   };

   std::ostream& operator << ( std::ostream& out, const mass_distribution& m );



   class rigid : public state 
   {

      mass_distribution distr;
      linalg::matrix inertia_inv; 

   public: 
      // xxxx,vvvv is the state of the origin of the internal
      // coordinates. Alternatively one can say that xxxx transforms 
      // internal to external coordinates. vvvv has the property: 
      // Let P = xxxx(0,0,0).
      // f(x) = vvvv( x - P ) is the speed function of the rigid object
      // in external coordinates. This looks crazy, but is in fact 
      // natural. vvvv.v0 is the speed of the current position of (0,0,0)
      // of the rigid object, in external coordinates.
      // vvvv.omega is the angular speed in external coordinates.

      rigid( const linalg::rototrans& xxxx,
             const linalg::spinspeed& vvvv,
             const mass_distribution& distr );

      force_torque 
      effect_external( const linalg::vector& f,
                       const linalg::vector& acting_point ) const;
         // f and acting_point are in external coordinates.
         // Result is in external coordinates, and torque is relative to
         // mass center. 

      force_torque 
      effect_internal( const linalg::vector& f,
                       const linalg::vector& acting_point ) const;
         // f and acting point are in internal coordinates.
         // Result is in external coordinates, and torque is relative to
         // mass center. 

      linalg::rototrans int2ext( ) const;
      linalg::rototrans ext2int( ) const;
         // User should not look at state. Only use these functions.

      linalg::spinspeed intspeed( ) const;
         // Maps internal coordinates to speed in internal coordinates.
         // One should not write the same for external coordinates, due to
         // problems with floating point accuracy.

      linalg::vector speed2int( linalg::vector speed ) const
            { return xxxx. q. inverse( ). rotate( speed ); }
      linalg::vector speed2ext( linalg::vector speed ) const
            { return xxxx. q. rotate( speed ); } 
         // Fact that I have to include the word 'speed' in the name
         // is evidence of the fact that the type system is not rich enough.

      state_acc accelerations( const force_torque& ft ) const; 
         // Compute the linear and angular accelerations. 
         // ft must have force in external coordinates, 
         // torque relative to mass center. Values returned by
         // effect_internal/effect_external are suitable. 

      friend std::ostream& operator << ( std::ostream& , const rigid& ); 
   };

}

#endif 


