
#include <iostream>
#include <stdio.h>

#include <SFML/System.hpp>
#include <SFML/Window.hpp>
#include <SFML/OpenGL.hpp>

#include "linalg/tests.h" 
#include "linalg/fourmatrix.h"
#include "linalg/affinemap.h"
#include "linalg/rototrans.h"
#include "linalg/spinspeed.h"
#include "linalg/normdef.h" 
#include "linalg/generalvector.h"
#include "linalg/bandmatrix.h"
#include "linalg/generalmatrix.h"

#include "units.h"
#include "plotwindow.h"
#include "timer.h" 
#include "airfoil_mach0.h"
#include "twodimensional.h"
#include "logger.h"

#include "math/runge_kutta.h"
#include "math/constants.h" 
#include "math/complex.h"
#include "math/interpolator.h"

#include "graphics.h"

int main( int argc, char* argv [ ] )
{
   // Two dimensional plane simulation:

   logger altitude_log( 0.1 );
   logger speed_log( 0.1 );
   logger attitude_log( 0.1 );
   
   twodimensional b737( "B737-300", 51710.0, 2027731.0, 1.225,
      { 0, 1000, 0 }, 5.0_deg,    // position and attitude in LRU.
      { 90, 0,0 }, 0.2,         // speed and omega in LRU. 
      airfoils::fuselage( ),
      { 16, 0, 0 }, 
      { 12, 128, 0 }, 32,         // Fuselage surfaces, fuselage length.
      airfoils::naca4412( ), 
      { 2, -1, 0 }, 3.0_deg,      // Wing position, wing theta.
      103.0, 3.6,                 // Wing surface, wing chord.
      { 2, -2, 0 },
      0.1, 0.1,                     // Engine throttle, engine power.
      airfoils::naca0009( ),           // Aileron undeflected.
      airfoils::naca4412withflap( ),   // Aileron deflected.
      -0.0,                             // Current deflection.
      { -15, 1, 0 }, -0.6_deg,    // Elevator position, elevator theta. 
      33, 2.6 );                   // Elevator surface, elevator chord.
    
   double simtime = 0.0;
   double nextplot = simtime;
   double nextprint = simtime;

   plotwindow w( "Boeing 737-200 (red = altitude, green = attitude, blue = speed)",
      simtime - 120.0, simtime,
      -1.0, 1.0, 1800, 512 );

   timer tim( simtime, 10 );

   int state = 0;
   while( true )
   {
      if( state == 0 && simtime >= 300.0 )
      {
         b737. theta += 4.0_deg;
         state = 1;
      }

      if( state == 1 && simtime >= 600.0 )
      {
         // b737. throttle = 1.0;
         b737. elevator_theta -= 0.5_deg;
         state = 2;
      }

      if( simtime >= nextplot )
      {
         if( ! w. handle_events( nullptr ))
         {
             return 0;
         }
         w. setcenter( simtime - w. xhalfsize, 0.0 );

         w. prepare( );
        
         for( int s = -1; s <= 1; s += 2 )
         {
            w. plot( [ &altitude_log, s ] 
                     ( double x ) { return altitude_log( x, s ); },  
                        scenery::color(1,0,0)); 
            w. plot( [ &attitude_log, s ] 
                     ( double x ) { return attitude_log( x, s ); }, 
                     scenery::color(0,1,0)); 
            w. plot( [ &speed_log, s ]
                     ( double x ) { return speed_log( x, s ); },
                     scenery::color(0,0,1));
         }

         w. display( );

         altitude_log. erase( w. xcenter - w. xhalfsize );
         attitude_log. erase( w. xcenter - w. xhalfsize );
         speed_log. erase( w. xcenter - w. xhalfsize );

         nextplot += 0.1 * tim. getscale( );
      } 
      if( simtime >= nextprint )
      {
         std::cout << "-------------------------------\n";
         std::cout << "T = " << simtime << "\n";
         std::cout << b737;
         if( b737. position. y < 0 )
            std::cout << "(note that we are under the ground)\n"; 
         std::cout << "\n";

         std::cout << altitude_log. size( ) << "\n";
         
      }

      force_effect eg = b737. gravity( );
      force_effect ef = b737. fuselage( );
      force_effect emw = b737. main_wings( );
      force_effect ee = b737. engines( );
      force_effect eel = b737. elevators( );

      force_effect total = eg + ef + emw + ee + eel;

      if( simtime >= nextprint )
      {  
         std::cout << "\n";
         std::cout << "gravity: " << eg << "\n";
         std::cout << "fuselage: " << ef << "\n";
         std::cout << "main wings: " << emw << "\n";
         std::cout << "engines: " << ee << "\n";
         std::cout << "elevators: " << eel << "\n";
         std::cout << "total: " << total << "\n\n";
      }

      // Determine step time:

      double h = tim. gettime( ) - simtime;
      if( h > 10.0 && tim. gettime( ) > 1.0 )
         throw std::runtime_error( "we lost the time" );

      if( h > 1.0E-4 ) h = 1.0E-4;

      // Integration:
      // This is an application of Euler's method, as described in the 
      // slides. 

      linalg::vector linacc = 
         { total.linear.x / b737. mass /* + b737.omega * b737.v0.y */,  
           total.linear.y / b737. mass /* - b737.omega * b737.v0.x */, 
           0 };

      double angacc = total.torque / b737. inertia; 

      if( simtime >= nextprint )
      {
         std::cout << "linear acceleration = " << linacc << "\n";
         std::cout << "angular acceleration = " << angacc << "\n";
         std::cout << "\n";
      }

      linalg::vector newv0 = b737.v0 + h * linacc;
      double newomega = b737.omega + h * angacc;

      b737. position += 0.5 * h * ( b737. v0 + newv0 );
      b737. theta += 0.5 * h * ( b737. omega + newomega );

      b737. v0 = newv0;
      b737. omega = newomega;

      b737. engine_power += 0.4 * h * ( b737. throttle - b737. engine_power );

      altitude_log. insert( simtime, b737. position. y / 2000 );
      speed_log. insert( simtime, b737. v0. length( ) / 100.0 );
      attitude_log. insert( simtime, b737. theta / 10.0_deg );

      while( simtime >= nextprint )
         nextprint += 1.0; 

      simtime += h; 
   }

}


